#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <libintl.h>
#include <locale.h>

#include "tampon.h"

#define TLL_ETQ 4096  /* taille initiale du tampon à étiquettes */
#define TLL_CTN 4096  /* taille initiale du tampon à texte */

#define DRAP_OUV 01   /* parenthèse ouvrante */
#define DRAP_FRM 02   /* parenthèse fermante */

#define NGUI_BLOC 3   /* nombre de guillemets pour ouvrir ou fermer un bloc */

#define TP_TYPE_ETQ 0 /* tampon contenant des étiquettes */
#define TP_TYPE_CTN 1 /* tampon contenant du « contenu » */

#define JT_OUV 0      /* parenthèse ouvrante */
#define JT_FRM 1      /* parenthèse fermante */
#define JT_FCT 2      /* séparateur fictif */
#define JT_ATM 3      /* atome */

#define  _(c) gettext(c)

static void utilisation(void);
static int  transformer(FILE *);
static int  prch_jeton(FILE *, struct tampon *, int);
static int  afficher_ligne(struct tampon *, struct tampon *, int *);
static int  guillemets(FILE *, struct tampon *);
static int  bloc(FILE *, struct tampon *);

char *nom_prog;       /* nom du programme */

struct tampon etq;    /* tampon à étiquettes */
struct tampon ctn;    /* tampon à texte */

int
main(int argc, char **argv)
{
	if ((nom_prog = argv[0]) == NULL || nom_prog[0] == '\0')
		nom_prog = "aplat";

#ifdef __OpenBSD__
	if (unveil(I18N_DOS, "r") < 0 ||
			unveil("/usr/share/locale/UTF-8/LC_CTYPE", "r") < 0) {
		fprintf(stderr, _("%s: erreur avec unveil()\n"), nom_prog);
		return 1;
	}
	if (pledge("stdio rpath", NULL) < 0) {
		fprintf(stderr, _("%s: erreur avec pledge()\n"), nom_prog);
		return 1;
	}
#endif

	setlocale(LC_ALL, "");
	bindtextdomain(I18N_DOMAINE, I18N_DOS);
	textdomain(I18N_DOMAINE);

	if (argc > 1) {
		utilisation();
		return 1;
	}

	/* initialiser les tampons */
	if (tp_init(&etq, TLL_ETQ) < 0 || tp_init(&ctn, TLL_CTN) < 0)
		return 1;

	if (transformer(stdin) < 0)
		return 1;

	return 0;
}

static void
utilisation(void)
{
	printf(_("Utilisation: %s <[fichier]\n"), nom_prog);
}

static int transformer(FILE *f)
{
	int prfd;  /* profondeur */
	int draps; /* drapeaux */
	int c;

	assert(f != NULL);

	prfd  = 0;
	draps = 0;

	switch (prch_jeton(f, &etq, TP_TYPE_ETQ)) {
	case JT_OUV:
		break;
	case EOF:
		return 0;
	default:
		fprintf(stderr,
			_("%s: le premier caractère non blanc "
			  "doit être « ( »\n"),
			nom_prog);
		return -1;
	}

ouv:
	prfd++;

	tp_ecr(':', &etq);

	switch (prch_jeton(f, &etq, TP_TYPE_ETQ)) {
	case JT_OUV:
		draps |= DRAP_OUV;
		afficher_ligne(&etq, &ctn, &draps);
		goto ouv;
	case JT_FRM:
		draps |= DRAP_OUV | DRAP_FRM;
		afficher_ligne(&etq, &ctn, &draps);
		goto ferm;
	case JT_FCT:
		draps |= DRAP_OUV;
		afficher_ligne(&etq, &ctn, &draps);
		tp_ecr('\\', &etq);
		goto ouv;
	case JT_ATM:
		draps |= DRAP_OUV;
		goto ctn;
	case EOF:
		goto deseq;
	}
	assert(0);

ctn:
	switch (prch_jeton(f, &ctn, TP_TYPE_CTN)) {
	case JT_OUV:
		afficher_ligne(&etq, &ctn, &draps);
		goto ouv;
	case JT_FRM:
		draps |= DRAP_FRM;
		afficher_ligne(&etq, &ctn, &draps);
		goto ferm;
	case JT_ATM:
		goto ctn;
	case EOF:
		goto deseq;
	}
	assert(0);

ferm:
	if (--prfd == 0)
		goto fin;
	assert(prfd > 0);

	while (tp_etq_rec(&etq) != ETQ_SEP_REL) {
		draps |= DRAP_FRM;
		afficher_ligne(&etq, &ctn, &draps);
		if (--prfd == 0)
			goto fin;
	}

	switch (prch_jeton(f, &ctn, TP_TYPE_CTN)) {
	case JT_OUV:
		goto ouv;
	case JT_FRM:
		draps |= DRAP_FRM;
		afficher_ligne(&etq, &ctn, &draps);
		goto ferm;
	case JT_ATM:
		goto ctn;
	case EOF:
		goto deseq;
	}
	assert(0);

fin:
	while ((c = getc(f)) == ' ' || c == '\n' || c == '\t')
		;
	if (c == EOF)
		return 0;
	fprintf(stderr, _("%s: aucun caractère non blanc ne peut suivre "
			  "le dernier « ) »\n"),
		nom_prog);
	return -1;

deseq:
	fprintf(stderr, _("%s: parenthèses déséquilibrées\n"), nom_prog);
	return -1;
}

static int
prch_jeton(FILE *f, struct tampon *tp, int tp_type)
{
	int c;
	int ngui; /* nombre de guillemets */

	assert(f != NULL);
	assert(tp != NULL);
	assert(tp_type == TP_TYPE_ETQ || tp_type == TP_TYPE_CTN);

	while ((c = getc(f)) == ' ' || c == '\n' || c == '\t')
		if (c == EOF)
			return EOF;

	switch (c) {
	case '(':
		return JT_OUV;
	case ')':
		return JT_FRM;
	}

	do {
		switch (c) {
		case '(': case ')':
			ungetc(c, f);
			/* CASCADE */
		case ' ': case '\t': case '\n':
			return JT_ATM;
		case '"':
			/* compter le nombre de guillemets */
			for (ngui = 1; (c = getc(f)) == '"'; ngui++)
				if (c == EOF)
					return EOF;
			if (ngui == NGUI_BLOC) {
				/* ignorer le reste de la ligne */
				while (c != '\n')
					if ((c = getc(f)) == EOF)
						return EOF;
				if (bloc(f, tp) == EOF)
					return EOF;
				continue;
			}
			ungetc(c, f);
			if (ngui % 2 == 1) { /* nombre impaire de guillemets */
				if (guillemets(f, tp) == EOF)
					return EOF;
			}
			continue;
		case ':':
			if (tp_type == TP_TYPE_ETQ)
				return JT_FCT;
			tp_ecr(':', tp);
		case '\\':
			switch (c = getc(f)) {
			case '\\':
				tp_ecr('\\', tp);
				break;
			case '\t':
				tp_ecr('\\', tp);
				tp_ecr('t', tp);
				continue;
			case '\n':
				continue;
			case EOF:
				return EOF;
			}
			/* CASCADE */
		default:
			tp_ecr(c, tp);
		}
	} while ((c = getc(f)) != EOF);
	return EOF;
}

static int
afficher_ligne(struct tampon *etq, struct tampon *ctn, int *draps)
{
	char *c;

	assert(etq     != NULL);
	assert(ctn     != NULL);
	assert(draps   != NULL);
	assert(etq->tp != NULL);
	assert(ctn->tp != NULL);
	assert(etq->pc != NULL);
	assert(ctn->pc != NULL);

	/* étiquettes */
	for (c = etq->tp; c < etq->pc; c++) {
		if (*c == '\\')
			/* les autres /\\./ doivent être préservés */
			if (*++c != ':')
				putc('\\', stdout);
		if (putc(*c, stdout) < 0)
			return -1;
	}
	putc(':', stdout); /* ajouter le point final */
	putc('\t', stdout);

	/* drapeaux */
	if (*draps & DRAP_OUV)
		putc('(', stdout);
	if (*draps & DRAP_FRM)
		putc(')', stdout);
	*draps = 0;

	putc('\t', stdout);

	/* contenu */
	for (c = ctn->tp; c < ctn->pc; c++)
		if (putc(*c, stdout) < 0)
			return -1;

	tp_eff(ctn);
	putc('\n', stdout);
	return 0;
}

static int
guillemets(FILE *f, struct tampon *tp)
{
	int c;

	assert(f != NULL);
	assert(tp != NULL);

	while ((c = getc(f)) != '"' && c != EOF) {
		switch (c) {
		case '\n':
			tp_ecr('\\', tp);
			tp_ecr('n', tp);
			continue;
		case '\t':
			tp_ecr('\\', tp);
			tp_ecr('t', tp);
			continue;
		case '\\':
			switch (c = getc(f)) {
			case '\\':
				tp_ecr('\\', tp);
				break;
			case '\n':
				continue;
			case '\t':
				tp_ecr('\\', tp);
				tp_ecr('t', tp);
				continue;
			case EOF:
				return EOF;
			}
			/* CASCADE */
		default:
			tp_ecr(c, tp);
		}
	}
	return c;
}

static int
bloc(FILE *f, struct tampon *tp)
{
	int    c;
	int    ngui;
	size_t ncar;

	assert(f != NULL);
	assert(tp != NULL);

	ncar = tp_ncar(tp);

	while ((c = getc(f)) != EOF) {
		switch (c) {
		case '\n':
			tp_ecr('\\', tp);
			tp_ecr('n', tp);
			continue;
		case '\t':
			tp_ecr('\\', tp);
			tp_ecr('t', tp);
			continue;
		case '"':
			/* compter le nombre de guillemets */
			for (ngui = 1; (c = getc(f)) == '"' && ngui < NGUI_BLOC;
					ngui++)
				;
			if (c != '!' || ngui < NGUI_BLOC)
				ungetc(c, f);
			if (ngui == NGUI_BLOC) {
				if (c != '!') {
					/* si le bloc n'est pas vide */
					if (ncar != tp_ncar(tp))
						tp_bloc_rec(tp);
					return c;
				}
			}
			while (ngui-- > 0)
				tp_ecr('"', tp);
			continue;
		case '\\':
			tp_ecr('\\', tp);
			/* CASCADE */
		default:
			tp_ecr(c, tp);
		}
	}
	return EOF;
}
